# global imports (dont't worry, red is normal)
import network_constants as netconst
from database.MongoRepositoryBase import MongoRepositoryBase
from db.entities.layer_adapter import LayerAdapter

import pymongo
import json
from typing import List

class Repository(MongoRepositoryBase):
    '''This is a repository for MongoDb.'''

    def __init__(self):
        super().__init__(netconst.BUSINESS_LOGIC_DB_HOSTNAME,
                         netconst.BUSINESS_LOGIC_DB_PORT,
                         'rest-gateway-db')

        self._adapter_collection = 'layer_adapters'

    def add(self, adapter : LayerAdapter):
        super().insert_entry(self._adapter_collection, adapter.to_serializable_dict())
    
    def one_by_name(self, name : str) -> LayerAdapter:
        return list(super().get_entries(self._adapter_collection, selection={"name": name}))

    def update(self, adapter : LayerAdapter):
        collection = self._database[self._adapter_collection]
        collection.update_one({"name":adapter.name}, {"$set": adapter.to_serializable_dict()})
    
    def delete_all_with_name(self, name: str):
        collection = self._database[self._adapter_collection]
        collection.delete_many({"name": name})

    def all(self) -> List[LayerAdapter]:
        result = super().get_entries(self._adapter_collection, projection={'_id': False})

        return list(result)