#global imports
from db.repository import Repository
from db.entities.layer_adapter import LayerAdapter

from typing import List


class LayerAdapterService:

    _repository = Repository()

    @staticmethod
    def all() -> List[LayerAdapter]:
        '''
        Return all currently defined layers
        '''
        return LayerAdapterService._repository.all()

    @staticmethod
    def update(layer: LayerAdapter):
        '''
        Overwrite the stored instance with the given one which is 
        identified by the layer name

        @params:
            layer - Required : layer object holding the current data
        '''
        LayerAdapterService._repository.update(layer)

    @staticmethod
    def add(name: str):
        '''
        Add a new layer to the DB. Attribute mapping and cluster
        attributes will be empty per default

        @params:
            name - Required : Unique name for a layer.
        '''
        adapter_new = LayerAdapter(name, {}, [])

        LayerAdapterService._repository.add(adapter_new)

    @staticmethod
    def add_complete(layer: LayerAdapter):
        '''
        Add a new layer to the DB. Attribute mappings and cluster
        attributes of the given layer are used

        @params:
            layer - Required : layer object holding correct data
        '''
        LayerAdapterService._repository.add(layer)

    @staticmethod
    def delete(layer: LayerAdapter):
        '''
        delete all layers with the given name.

        @params:
            layer - Required : layer object to remove from the DB
        '''
        LayerAdapterService._repository.delete_all_with_name(layer.name)

    @staticmethod
    def one(name: str) -> LayerAdapter:
        '''
        Retrieve a single layer from the DB. Returns None if no layer
        was found under this name.

         @params:
            name - Required : Unique name for a layer.
        '''
        result = LayerAdapterService._repository.one_by_name(name)

        if len(result) == 1:
            return LayerAdapter.from_serializable_dict(result[0])
        else:
            return None
